/* vim: set tabstop=8 softtabstop=2 shiftwidth=2 expandtab: */
/*
 *    Copyright (C) 2011 Fabian Deutsch
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License along
 *    with this program; if not, write to the Free Software Foundation, Inc.,
 *    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

using Gst;

extern void gst_rtsp_media_factory_set_auth (Gst.RTSPMediaFactory factory, 
                                             Gst.RTSPAuth auth);


namespace Media
{
  public void init (ref unowned string[] args)
  {
    Gst.init (ref args);
  }

  public class RTSPClient : GLib.Object
  {
    dynamic Gst.Element playbin;
    dynamic Gst.Element video_sink;
    dynamic Gst.Element audio_sink;

    public string uri { 
      get; 
      set; 
      default = "";
    }

    public bool mute { 
      get; 
      set; 
      default = false;
    }


    public RTSPClient (Gst.Element vsink)
    {
      this.video_sink = vsink;
      build_playbin ();
    }

    public RTSPClient.playing (Gst.Element vsink, string uri)
    {
      this (vsink);
      this.uri = uri;
      this.start ();
    }

    ~RTSPClient ()
    {
      this.stop ();
#if DEBUG
      debug ("RTSPClient destroyed.");
#endif
    }

    void build_playbin ()
    {      
      audio_sink = Gst.ElementFactory.make ("pulsesink", "ps");
      audio_sink.stream_properties = Gst.Structure.from_string ("props," 
                                        + "media.role=phone," 
                                        + "filter.want=echo-cancel", null);
      this.bind_property ("mute", audio_sink, "mute", BindingFlags.DEFAULT);


      playbin = Gst.ElementFactory.make ("playbin2", "pb");
      playbin.video_sink = video_sink;
      playbin.audio_sink = audio_sink.ref (); //FIXME fixed upstream floating-ref

      playbin.source_setup.connect ( (e)  => {
#if DEBUG
        debug ("Setting latency.");
#endif
        dynamic Gst.Object element = e;
        dynamic Gst.Object src = element.source;
        src.latency = 250; // FIXME accessing property results in segfault
        src.udp_buffer_size = 2 * 1024 * 1024;
      });
      // FIXME why doesn't the following line work?
      // this.bind_property ("uri", playbin, "uri", BindingFlags.SYNC_CREATE);
    }

    public void start ()
    {
      playbin.uri = this.uri;
#if DEBUG
      debug (@"State play on $(this.uri).");
      debug ("-> %s", playbin.uri);
#endif
      audio_sink.client = @"presence (out) from '$(this.uri)'";
      playbin.set_state (Gst.State.PLAYING);
    }
    
    public void stop ()
    {
      playbin.set_state (Gst.State.NULL);
    }
  }




  public class LocalRTSPServer : GLib.Object
  {
    MainLoop loop;

    Gst.RTSPServer server;
    Gst.RTSPMediaMapping mapping;

    /*
     * The actual video_size is calculated using this base sizes and the 
     * video_scale_factor.
     */
    double base_width = 1280;
    double base_height = 720;


    /*
     * The video_quality_factor is not applied to the whole range (as it doesn't
     * make sense for schroenc's quality), but a limited subset. The actual 
     * quality is calculated based on this lower and upper limits an the 
     * video_scale_factor.
     */
    double base_quality_min = 4;
    double base_quality_max = 9;

    /*
     * FPS factor
     */
    double base_fps_min = 4;
    double base_fps_max = 30;

    /*
     * Determins if the server should be bound to a IPv6 interface.
     */
    public bool use_ipv6 {
      get;
      set;
      default = false;
    }

    /*
     * What port to bind to (could also be a string, but we restrict it).
     */
    public string port {
      get;
      set;
      default = "8554";
    }

    /*
     * The mountpoint we bind a/v to.
     */
    public string mountpoint {
      get;
      set;
      default = "presence";
    }

    /*
     * The RTSP address we are bound to and where the a/v stream an be reached.
     */
    public string address {
      get;
      private set;
    }

    public string user {
      get;
      private set;
      default = "presence";
    }

    public string? password {
      get;
      set;
      default = null;
    }

    public string video_device {
      get;
      set;
      default = "/dev/video0";
    }

    /*
     * See above.
     * Note: This can not be changed while playing.
     */
    [FloatingType (min = 0, max = 1)]
    public double video_scale_factor {
      get;
      set;
      default = 0.25;
    }

    /*
     * The actual width used.
     */
    public uint video_width {
      get { return (uint) (this.base_width * this.video_scale_factor); }
      set { this._video_height_request = value; }
    }
    uint? _video_height_request = null;

    /*
     * The actual height used.
     */
    public uint video_height {
      get { return (uint) (this.base_height * this.video_scale_factor); }
      set { this._video_width_request = value; }
    }
    uint? _video_width_request = null;

    public string video_size {
      owned get { return @"$(this.video_width)x$(this.video_height)"; }
    }

    /*
     * See above.
     * Note: A change has no effect while playing. (Really?)
     */
    [FloatingType (min = 0, max = 1)]
    public double video_quality_factor {
      get;
      set;
      default = 0.75;
    }

    /*
     * The actual quality used.
     */
    public double video_quality {
      get { 
        double range = (this.base_quality_max - this.base_quality_min);
        return this.base_quality_min + range * this.video_quality_factor; 
      }
      set { this._video_quality_request = value; }
    }
    double? _video_quality_request = null;

    /*
     * See above.
     * Note: A change has no effect while playing. (Really?)
     */
    [FloatingType (min = 0, max = 1)]
    public double video_fps_factor {
      get;
      set;
      default = 0.5;
    }

    /*
     * The actual fps used.
     */
    public double video_fps {
      get { 
        double range = (this.base_fps_max - this.base_fps_min);
        return this.base_fps_min + range * this.video_fps_factor; 
      }
      set { this._video_fps_request = value; }
    }
    double? _video_fps_request = null;

    public LocalRTSPServer ()
    {
      server = new Gst.RTSPServer ();
      mapping = server.get_media_mapping ();

      this.bind_property ("port", this.server, "service", 
                          BindingFlags.SYNC_CREATE);
    }

    public void attach (MainLoop l)
    {
      this.loop = l;

      // Finalize server setup:
      server.address = use_ipv6 ? "::0" : "0.0.0.0";
      this.address = @"rtsp://$(server.address):$(port)/$(mountpoint)";

      // Add mappings:
      add_mapping (this.mountpoint, build_default_launchline ());

      server.attach (loop.get_context ());

#if DEBUG
      debug (@"Stream ready at $(this.address)");
#endif
    }

    public void add_mapping (string mountpoint, string launchline)
    {
      Gst.RTSPMediaFactory factory = new Gst.RTSPMediaFactory ();
      Gst.RTSPAuth auth = new Gst.RTSPAuth ();

      if (this.password != null)
      {
        auth.set_basic (Gst.RTSPAuth.make_basic (this.user, this.password));
        // FIXME Upstream binding
        //factory.set_auth (this.auth);
        gst_rtsp_media_factory_set_auth (factory, auth);
      }

      factory.shared = true;
      factory.set_launch (launchline);

      factory.ref ();  //FIXME needed because mapping  bindingis wrong.
      mapping.add_factory ("/" + mountpoint, factory);

      warning (@"Mapping added for '$(mountpoint)' with: $(launchline)");

    }

    string build_default_launchline ()
    {
      char[] qustr = new char[128]; // FIXME decimal point

      string launchline = """( 
  (
    v4l2src device=%s 
    ! videorate 
    ! aspectratiocrop aspect-ratio=16/9 
    ! videoscale 
    ! video/x-raw-yuv,width=%.0f,height=%.0f,framerate=%.0f/1 
    ! schroenc force-profile=vc2_simple quality=%s queue-depth=5
  )
  ! queue 
  ! rtpgstpay pt=96 name=pay0

  (
    pulsesrc do-timestamp=true stream-properties='props,media.role=phone,filter.want=echo-cancel' client='Presence (in)'
    ! vobisenc 
  )
  ! queue 
  ! rtpgstpay pt=97 name=pay1 
)""".printf (this.video_device,
             this.video_width, 
             this.video_height, 
             this._video_fps_request ?? this.video_fps,
             this.video_quality.to_str (qustr));

// launchline for cbr
//    ! schroenc force-profile=vc2_main rate-control=constant_bitrate bitrate=%d queue-depth=5
// rate=(int)(video_width * video_height * video_fps * 3));

// simplistic, w/o arith:
// schroenc force-profile=vc2_simple quality=%s queue-depth=5 

// simplistic w/ arith
// schroenc force-profile=vc2_main quality=%s queue-depth=5

      return launchline;
    }

    public void quit ()
    {
      if (loop != null)
      {
        loop.quit ();
      }
    }
  }
}
